typedef unsigned char opcode_t;
typedef unsigned char subcode_t;
typedef unsigned char ref_type_t;
typedef unsigned char addr_t;

typedef struct memory_ref_t {
  ref_type_t type;
  addr_t     addr;
} memory_ref_t;

#define SET_TYPE(type,value) ((type) = ((type)&0xf0)| ((value)&0x0f))
#define SET_MODE(type,value) ((type) = ((type)&0x0f)| (((value)&0x0f)<<4))
#define GET_MODE(type) (((type)>>4)&0x0f)
#define GET_TYPE(type) ((type)&0x0f)

#define CONSTANT      0
#define DIRECT        1
#define INDIRECT      2

#define DATA_SECTION  0
#define CODE_SECTION  1
#define STACK_SECTION 2
#define BSS_SECTION   3

typedef struct ret_t {
  memory_ref_t arg1;
} ret_t, jump_t, fork_t, exit_t, exec_t,
    sbrk_t, read_t, write_t, getppid_t;

typedef struct assign_t {
  memory_ref_t arg1;
  memory_ref_t arg2;
} assign_t, signal_t, kill_t, wait_t;

typedef struct if_t {
  memory_ref_t arg1;
  memory_ref_t arg2;
  memory_ref_t arg3;
} if_t, call_t;

typedef struct instruction_t {
  opcode_t opcode;
  union {
    assign_t  assign;
    if_t      si;
    jump_t    jump;
    call_t    call;
    ret_t     ret;
    fork_t    fork;
    exit_t    exit;
    exec_t    exec;
    wait_t    wait;
    signal_t  signal;
    kill_t    kill;
    sbrk_t    sbrk;
    read_t    read;
    write_t   write;
    getppid_t getppid;
  } datas;
  unsigned char padding;
} instruction_t;

#define UNK          -1
#define ASSIGN_PLUS  0
#define ASSIGN_MINUS 1
#define ASSIGN_MULT  2
#define ASSIGN       3
#define IF           4
#define JUMP         5
#define CALL         6
#define RET          7
#define FORK         8
#define EXEC         9
#define EXIT         10
#define WAIT         11
#define SIGNAL       12
#define KILL         13
#define READ         14
#define WRITE        15
#define GETPPID      16
#define LAST_OPCODE  16

#define U_SIGILL     0
#define U_SIGMEMOVER 1
#define U_SIGMEMWR   2
#define U_SIGUSER    3
#define U_SIGKILL    4

static unsigned char *opcodes[] = {
  /* Standard instruction set */
  "ASSIGN+", "ASSIGN-", "ASSIGN*", "ASSIGN", "IF", "JUMP", "CALL", "RET",
  /* System call set */
  "FORK", "EXEC", "EXIT", "WAIT", "SIGNAL", "KILL", "READ", "WRITE",
  "GETPPID"
};
