%{
#include <strings.h>
#include "assembler.h"
#include "consts.h"
#include "y.tab.h"
static int type;
static     char *text;
int        numligne = 1;
%}
integer    [0-9]+
memtype    [DBS]
eol        \n
separator  [ \t]+
label      [a-zA-Z_]+
%%
ASSIGN\+ {
  yylval.opcode = ASSIGN_PLUS;
  return INST_ASSIGN;
}
ASSIGN\- {
  yylval.opcode = ASSIGN_MINUS;
  return INST_ASSIGN;
}
ASSIGN\* {
  yylval.opcode = ASSIGN_MULT;
  return INST_ASSIGN;
}
ASSIGN {
  yylval.opcode = ASSIGN;
  return INST_ASSIGN;
}
IF {
  yylval.opcode = IF;
  return INST_IF;
}
JUMP {
  yylval.opcode = JUMP;
  return INST_JUMP;
}
CALL {
  yylval.opcode = CALL;
  return INST_CALL;
}
RET {
  yylval.opcode = RET;
  return INST_RET;
}
FORK {
  yylval.opcode = FORK;
  return INST_FORK;
}
EXEC {
  yylval.opcode = EXEC;
  return INST_EXEC;
}
EXIT {
  yylval.opcode = EXIT;
  return INST_EXIT;
}
WAIT {
  yylval.opcode = WAIT;
  return INST_WAIT;
}
SIGNAL {
  yylval.opcode = SIGNAL;
  return INST_SIGNAL;
}
KILL {
  yylval.opcode = KILL;
  return INST_KILL;
}
READ {
  yylval.opcode = READ;
  return INST_READ;
}
WRITE {
  yylval.opcode = WRITE;
  return INST_WRITE;
}
GETPPID {
  yylval.opcode = GETPPID;
  return INST_GETPPID;
}
U_SIGILL {
  SET_MODE(yylval.ref.type,CONSTANT);
  yylval.ref.addr = U_SIGILL;
  return CSTSIG;
}
U_SIGUSER {
  SET_MODE(yylval.ref.type,CONSTANT);
  yylval.ref.addr = U_SIGUSER;
  return CSTSIG;
}
U_SIGMEMOVER {
  SET_MODE(yylval.ref.type,CONSTANT);
  yylval.ref.addr = U_SIGMEMOVER;
  return CSTSIG;
}
U_SIGMEMWR {
  SET_MODE(yylval.ref.type,CONSTANT);
  yylval.ref.addr = U_SIGMEMWR;
  return CSTSIG;
}
U_SIGKILL {
  SET_MODE(yylval.ref.type,CONSTANT);
  yylval.ref.addr = U_SIGKILL;
  return CSTSIG;
}
\#{integer} {
  SET_MODE(yylval.ref.type,CONSTANT);
  yylval.ref.addr = atoi(yytext+1);
  return CST;
}
[\*]?{memtype}{integer}	{
  text = yytext;
  if (text[0]=='*') {
    SET_MODE(yylval.ref.type,INDIRECT);
    text++;
  }
  else {
    SET_MODE(yylval.ref.type,DIRECT);
  }
  switch(text[0]) {
  case 'D': type = DATA_SECTION;  break;
  case 'B': type = BSS_SECTION;   break;
  case 'S': type = STACK_SECTION; break;
  }
  SET_TYPE(yylval.ref.type,type);
  yylval.ref.addr = atoi(text+1);
  return MEM_REF;
}
{label} {
  if (strlen(yytext)>40) {
    fprintf(stderr,"Attention: etiquettes limitees a %d caracteres (%s)\n",
	    LABEL_MAX,yytext);
    strncpy(yylval.label,yytext,LABEL_MAX);
  }
  else {
    strcpy(yylval.label,yytext);
  }
  return LABEL;
}
{separator} {
  return SEPARATOR;
}
\/.* {
  numligne++;
  return COMMENT;
}
{eol} {
  numligne++;
  return EOL;
}
. {
  return yytext[0];
}
%%
int yyerror(const char *msg) {
  fprintf(stderr,"Erreur de syntaxe en ligne %d (%s)\n",numligne,yytext);
  exit(1);
}
