%{
#include "assembler.h"
#include "consts.h"
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
  int yyerror();
  int yylex();
  
  static int inlabels(const unsigned char *label);
  static void add_unsolved(unsigned char *label,opcode_t opcode);

  static instruction_t tab_ins[MAX_INS];
  static int current_instruction = 2;
  static int current_label = 0;
  static int current_unsolved_label = 0;
  static int l;
  static struct {
    unsigned char label[LABEL_MAX+1];
    int           instruction_num;
  } labels[MAX_LABELS];
  extern int numligne;
  static struct {
    unsigned char label[LABEL_MAX+1];
    opcode_t opcode;
    int instruction_num;
  } unsolved_labels[MAX_LABELS];
%}

%union {
  memory_ref_t  ref;
  opcode_t      opcode;
  unsigned char label[LABEL_MAX+1];
  instruction_t instruction;
}

%token <opcode> INST_ASSIGN
%token <opcode> INST_IF
%token <opcode> INST_JUMP
%token <opcode> INST_CALL
%token <opcode> INST_RET
%token <opcode> INST_FORK
%token <opcode> INST_EXEC
%token <opcode> INST_EXIT
%token <opcode> INST_WAIT
%token <opcode> INST_SIGNAL
%token <opcode> INST_KILL
%token <opcode> INST_READ
%token <opcode> INST_WRITE
%token <opcode> INST_GETPPID
%token SEPARATOR
%token EOL
%token COMMENT
%token <ref> MEM_REF
%token <ref> CST
%token <ref> CSTSIG
%token <label> LABEL
%type  <ref> mem_ref_or_cst
%type  <ref> mem_ref_or_cstsig
%type  <instruction> inst_assign
%type  <instruction> inst_if
%type  <instruction> inst_jump
%type  <instruction> inst_call
%type  <instruction> inst_ret
%type  <instruction> inst_fork
%type  <instruction> inst_exec
%type  <instruction> inst_exit
%type  <instruction> inst_wait
%type  <instruction> inst_signal
%type  <instruction> inst_kill
%type  <instruction> inst_read
%type  <instruction> inst_write
%type  <instruction> inst_getppid
%type  <instruction> inst
%%
insts_list :
	   inst_or_label_or_comment EOL
	 | insts_list inst_or_label_or_comment EOL
	 ;

inst_or_label_or_comment :
	   SEPARATOR inst { tab_ins[current_instruction++] = $2; }
	 | label ':'
	 | comment
	 ;
comment : COMMENT
label :
LABEL {
  if (inlabels($1)==-1) {
    strcpy(labels[current_label].label,$1);
    labels[current_label++].instruction_num = current_instruction;
  }
  else {
    fprintf(stderr,"Redefinition de label en ligne %d (%s)\n",
	    numligne,$1);
    exit(1);
  }
}
;
inst :
           inst_assign
	 | inst_if
	 | inst_jump
	 | inst_call
	 | inst_ret
         | inst_fork
         | inst_exec
         | inst_exit
         | inst_wait
         | inst_signal
         | inst_kill
         | inst_read
         | inst_write
         | inst_getppid
	 ;

inst_assign :
	   INST_ASSIGN SEPARATOR MEM_REF ',' mem_ref_or_cst
		{ $$.opcode = $1;
		  $$.datas.assign.arg1 = $3;
		  $$.datas.assign.arg2 = $5; }
         ;
inst_if :
	   INST_IF SEPARATOR mem_ref_or_cst ',' mem_ref_or_cst ',' LABEL
		{ $$.opcode = $1;
		  $$.datas.si.arg1 = $3;
		  $$.datas.si.arg2 = $5;
		  if ((l=inlabels($7))!=-1) {
		    SET_TYPE($$.datas.si.arg3.type,CODE_SECTION);
		    $$.datas.si.arg3.addr = labels[l].instruction_num;
		  }
		  else {
		    add_unsolved($7,$1);
		  }
		}
         ;
inst_fork :
INST_FORK SEPARATOR MEM_REF {
  $$.opcode = $1;
  $$.datas.fork.arg1 = $3;
}
;
inst_exec :
INST_EXEC SEPARATOR mem_ref_or_cst {
  $$.opcode = $1;
  $$.datas.exec.arg1 = $3;
}
;
inst_exit :
INST_EXIT SEPARATOR mem_ref_or_cst {
  $$.opcode = $1;
  $$.datas.exit.arg1 = $3;
}
;
inst_wait :
INST_WAIT SEPARATOR MEM_REF ',' MEM_REF {
  $$.opcode = $1;
  $$.datas.wait.arg1 = $3;
  $$.datas.wait.arg2 = $5;
}
;
inst_signal :
INST_SIGNAL SEPARATOR mem_ref_or_cstsig ',' LABEL {
  $$.opcode = $1;
  $$.datas.signal.arg1 = $3;
  if ((l=inlabels($5))!=-1) {
    SET_TYPE($$.datas.signal.arg2.type,CODE_SECTION);
    $$.datas.signal.arg2.addr = labels[l].instruction_num;
  }
  else {
    add_unsolved($5,$1);
  }
}
;
inst_getppid :
INST_GETPPID SEPARATOR MEM_REF {
  $$.opcode = $1;
  $$.datas.getppid.arg1 = $3;
}
inst_kill :
INST_KILL SEPARATOR mem_ref_or_cstsig ',' MEM_REF {
  $$.opcode = $1;
  $$.datas.kill.arg1 = $3;
  $$.datas.kill.arg2 = $5;
}
;
inst_read :
INST_READ SEPARATOR MEM_REF {
  $$.opcode = $1;
  $$.datas.read.arg1 = $3;
}
;
inst_write :
INST_WRITE SEPARATOR mem_ref_or_cst {
  $$.opcode = $1;
  $$.datas.write.arg1 = $3;
}
;
inst_jump :
	  INST_JUMP SEPARATOR LABEL
		{ if ((l=inlabels($3))!=-1) {
		    SET_TYPE($$.datas.jump.arg1.type,CODE_SECTION);
		    $$.datas.jump.arg1.addr = labels[l].instruction_num;
		  } else {
		    add_unsolved($3,$1);
		  }
		}
         ;
inst_call :
	  INST_CALL SEPARATOR LABEL
		{ if ((l=inlabels($3))!=-1) {
		    SET_TYPE($$.datas.call.arg1.type,CODE_SECTION);
		    $$.datas.call.arg1.addr = labels[l].instruction_num;
		  } else {
		    add_unsolved($3,$1);
		  }
		}
	 ;
inst_ret :
	  INST_RET { $$.opcode = $1; }
         ;
mem_ref_or_cst :
	   MEM_REF
         | CST
         ;
mem_ref_or_cstsig :
	   MEM_REF
         | CSTSIG
         ;
%%
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

static int inlabels(const unsigned char *label) {
  int i;

  for (i=0; i<current_label; i++)
    if (!strcmp(label,labels[i].label)) return i;
  return -1;
}

static void add_unsolved(unsigned char *label,opcode_t opcode) {
  if (current_unsolved_label==MAX_LABELS) {
    fprintf(stderr,"Impossible de resoudre les etiquettes en avant\n");
    exit(1);
  }
  strcpy(unsolved_labels[current_unsolved_label].label,label);
  unsolved_labels[current_unsolved_label].opcode = opcode;
  unsolved_labels[current_unsolved_label++].instruction_num = current_instruction;
}

int main(int argc,char *argv[]) {
    int i, l, d, o;
    instruction_t *ins;

    if (argc!=3) {
      fprintf(stderr,"usage: %s <source> <objet>\n",argv[0]);
      exit(EXIT_FAILURE);
    }
    d = open(argv[1],O_RDONLY);
    if (d==-1) {
      fprintf(stderr,"%s: impossible d'ouvrir %s\n",argv[0],argv[1]);
      exit(EXIT_FAILURE);
    }
    close(0);
    dup(d);
    close(d);

    o = open(argv[2],O_WRONLY|O_CREAT|O_TRUNC,0644);
    if (o==-1) {
      fprintf(stderr,"%s: impossible d'ouvrir %s\n",argv[0],argv[2]);
      exit(EXIT_FAILURE);
    }

    yyparse();

    for (i=0; i<current_unsolved_label; i++) {
      if ((l=inlabels(unsolved_labels[i].label))==-1) {
	fprintf(stderr,"Label inconnu %s\n",unsolved_labels[i].label);
	exit(1);
      }
      ins = tab_ins+unsolved_labels[i].instruction_num;
      switch(unsolved_labels[i].opcode) {
      case CALL:
	SET_TYPE(ins->datas.call.arg1.type,CODE_SECTION);
	ins->datas.call.arg1.addr = labels[l].instruction_num;
	break;
      case JUMP:
	SET_TYPE(ins->datas.jump.arg1.type,CODE_SECTION);
	ins->datas.jump.arg1.addr = labels[l].instruction_num;
	break;
      case IF:
	SET_TYPE(ins->datas.si.arg3.type,CODE_SECTION);
	ins->datas.si.arg3.addr = labels[l].instruction_num;
	break;
      case SIGNAL:
	SET_TYPE(ins->datas.signal.arg2.type,CODE_SECTION);
	ins->datas.signal.arg2.addr = labels[l].instruction_num;
	break;
      default:
	fprintf(stderr,"Panique: voir avec JBY\n");
	exit(EXIT_FAILURE);
      }
    }
    /* TEST: do not use for real
       tab_ins[current_instruction++].opcode = UNK;
    */
    tab_ins[0].opcode = CALL;
    SET_TYPE(tab_ins[0].datas.jump.arg1.type,CODE_SECTION);
    if ((l=inlabels("debut"))==-1) {
      fprintf(stderr,"Attention il manque l'etiquette de debut\n");
      fprintf(stderr,"  ce code demarre donc a la premiere instruction!\n");
      tab_ins[0].datas.jump.arg1.addr = 2;
    }
    else {
      tab_ins[0].datas.jump.arg1.addr = labels[l].instruction_num;
    }
    tab_ins[1].opcode = EXIT;
    SET_MODE(tab_ins[1].datas.exit.arg1.type,CONSTANT);
    tab_ins[1].datas.exit.arg1.addr = 0;

    for (i=0; i<current_instruction; i++) {
      if (write(o,tab_ins+i,sizeof(instruction_t))!=sizeof(instruction_t)) {
	fprintf(stderr,"%s: erreur d'ecriture dans %s\n",argv[0],argv[2]);
      }
    }
    close(o);
    fprintf(stdout,"%s a ete correctement compile dans %s\n",argv[1],argv[2]);
    exit(EXIT_SUCCESS);
}
